@@ -37,7 +37,6 @@ import shlex
 import sys
 import threading
 import time
-import uuid
 
 from eventlet.green import subprocess as e_subprocess
 from eventlet import semaphore
@@ -45,6 +44,7 @@ from eventlet import timeout as e_timeout
 from oslo_config import cfg
 from oslo_log import log as logging
 from oslo_utils import excutils
+from oslo_utils import uuidutils
 import paramiko
 import requests
 from requests import adapters
@@ -56,6 +56,7 @@ from sahara.i18n import _
 from sahara.i18n import _LE
 from sahara.service import trusts
 from sahara.utils import crypto
+from sahara.utils import network as net_utils
 from sahara.utils.openstack import neutron
 from sahara.utils import procutils
 from sahara.utils import remote
@@ -172,7 +173,7 @@ def _execute_command(cmd, run_as_root=False, get_stderr=False,
     else:
         chan.exec_command(cmd)
 
-    # todo(dmitryme): that could hang if stderr buffer overflows
+    # TODO(dmitryme): that could hang if stderr buffer overflows
     stdout = _read_paramimko_stream(chan.recv)
     stderr = _read_paramimko_stream(chan.recv_stderr)
 
@@ -277,7 +278,7 @@ def _append_fl(sftp, remote_file, data):
 
 def _write_file(sftp, remote_file, data, run_as_root):
     if run_as_root:
-        temp_file = 'temp-file-%s' % six.text_type(uuid.uuid4())
+        temp_file = 'temp-file-%s' % uuidutils.generate_uuid()
         _write_fl(sftp, temp_file, data)
         _execute_command(
             'mv %s %s' % (temp_file, remote_file), run_as_root=True)
@@ -287,7 +288,7 @@ def _write_file(sftp, remote_file, data, run_as_root):
 
 def _append_file(sftp, remote_file, data, run_as_root):
     if run_as_root:
-        temp_file = 'temp-file-%s' % six.text_type(uuid.uuid4())
+        temp_file = 'temp-file-%s' % uuidutils.generate_uuid()
         _write_fl(sftp, temp_file, data)
         _execute_command(
             'cat %s >> %s' % (temp_file, remote_file), run_as_root=True)
@@ -338,7 +339,7 @@ def _read_file_from(remote_file, run_as_root=False):
 
     fl = remote_file
     if run_as_root:
-        fl = 'temp-file-%s' % (six.text_type(uuid.uuid4()))
+        fl = 'temp-file-%s' % (uuidutils.generate_uuid())
         _execute_command('cp %s %s' % (remote_file, fl), run_as_root=True)
 
     try:
@@ -353,6 +354,71 @@ def _read_file_from(remote_file, run_as_root=False):
                 'rm %s' % fl, run_as_root=True, raise_when_error=False)
 
 
+def __get_python_to_execute():
+    try:
+        _execute_command('python3 --version')
+    except Exception:
+        _execute_command('python2 --version')
+        return 'python2'
+    return 'python3'
+
+
+def _get_os_distrib():
+    python_version = __get_python_to_execute()
+    return _execute_command(
+        ('printf "import platform\nprint(platform.linux_distribution('
+         'full_distribution_name=0)[0])" | {}'.format(python_version)),
+        run_as_root=False)[1].lower().strip()
+
+
+def _get_os_version():
+    python_version = __get_python_to_execute()
+    return _execute_command(
+        ('printf "import platform\nprint(platform.linux_distribution()[1])"'
+         ' | {}'.format(python_version)), run_as_root=False)[1].strip()
+
+
+def _install_packages(packages):
+    distrib = _get_os_distrib()
+    if distrib == 'ubuntu':
+        cmd = 'RUNLEVEL=1 apt-get install -y %(pkgs)s'
+    elif distrib == 'fedora':
+        fversion = _get_os_version()
+        if fversion >= 22:
+            cmd = 'dnf install -y %(pkgs)s'
+        else:
+            cmd = 'yum install -y %(pkgs)s'
+    elif distrib in ('redhat', 'centos'):
+        cmd = 'yum install -y %(pkgs)s'
+    else:
+        raise ex.NotImplementedException(
+            _('Package Installation'),
+            _('%(fmt)s is not implemented for OS %(distrib)s') % {
+                'fmt': '%s', 'distrib': distrib})
+    cmd = cmd % {'pkgs': ' '.join(packages)}
+    _execute_command(cmd, run_as_root=True)
+
+
+def _update_repository():
+    distrib = _get_os_distrib()
+    if distrib == 'ubuntu':
+        cmd = 'apt-get update'
+    elif distrib == 'fedora':
+        fversion = _get_os_version()
+        if fversion >= 22:
+            cmd = 'dnf clean all'
+        else:
+            cmd = 'yum clean all'
+    elif distrib in ('redhat', 'centos'):
+        cmd = 'yum clean all'
+    else:
+        raise ex.NotImplementedException(
+            _('Repository Update'),
+            _('%(fmt)s is not implemented for OS %(distrib)s') % {
+                'fmt': '%s', 'distrib': distrib})
+    _execute_command(cmd, run_as_root=True)
+
+
 def _replace_remote_string(remote_file, old_str, new_str):
     old_str = old_str.replace("\'", "\''")
     new_str = new_str.replace("\'", "\''")
@@ -623,6 +689,17 @@ class InstanceInteropHelper(remote.Remote):
         gateway_host = None
         gateway_image_username = None
         if proxy_gateway_node and not host_ng.is_proxy_gateway:
+
+            # tmckay-fp in other words, if we are going to connect
+            # through the proxy instead of the node we are actually
+            # trying to reach
+
+            # okay, the node group that supplies the proxy gateway
+            # must have fps, but if a proxy is used the other
+            # nodes are not required to have an fp.
+
+            # so, this instance is assumed not to have a floating
+            # ip and we are going to get to it through the proxy
             access_instance = proxy_gateway_node
             gateway_host = proxy_gateway_node.management_ip
             ng = proxy_gateway_node.node_group
@@ -632,7 +709,16 @@ class InstanceInteropHelper(remote.Remote):
         if CONF.proxy_command:
             # Build a session through a user-defined socket
             proxy_command = CONF.proxy_command
-        elif CONF.use_namespaces and not CONF.use_floating_ips:
+
+        # tmckay-fp we have the node_group for the instance right here
+        # okay, this test here whether access_instance.management_ip is an
+        # fp -- just compare to internal?
+        # in the neutron case, we check the node group for the
+        # access_instance and look for fp
+        # in the nova case, we compare management_ip to internal_ip or even
+        # use the nova interface
+        elif CONF.use_namespaces and not net_utils.has_floating_ip(
+                access_instance):
             # Build a session through a netcat socket in the Neutron namespace
             proxy_command = (
                 'ip netns exec qrouter-{router_id} nc {host} {port}')
@@ -664,20 +750,24 @@ class InstanceInteropHelper(remote.Remote):
         finally:
             procutils.shutdown_subprocess(proc, _cleanup)
 
-    def _run_with_log(self, func, timeout, *args, **kwargs):
+    def _run_with_log(self, func, timeout, description, *args, **kwargs):
         start_time = time.time()
+
         try:
-            with e_timeout.Timeout(timeout, ex.TimeoutException(timeout)):
+            with e_timeout.Timeout(timeout,
+                                   ex.TimeoutException(timeout,
+                                                       op_name=description)):
                 return self._run(func, *args, **kwargs)
         finally:
-            self._log_command('%s took %.1f seconds to complete' % (
-                func.__name__, time.time() - start_time))
+            self._log_command('"%s" took %.1f seconds to complete' % (
+                description, time.time() - start_time))
 
-    def _run_s(self, func, timeout, *args, **kwargs):
+    def _run_s(self, func, timeout, description, *args, **kwargs):
         timeout = _get_ssh_timeout(func, timeout)
         _acquire_remote_semaphore()
         try:
-            return self._run_with_log(func, timeout, *args, **kwargs)
+            return self._run_with_log(func, timeout,
+                                      description, *args, **kwargs)
         finally:
             _release_remote_semaphore()
 
@@ -705,7 +795,17 @@ class InstanceInteropHelper(remote.Remote):
         if CONF.proxy_command:
             # Build a session through a user-defined socket
             proxy_command = CONF.proxy_command
-        elif info or (CONF.use_namespaces and not CONF.use_floating_ips):
+
+        # tmckay-fp again we can check the node group for the instance
+        # what are the implications for nova here? None, because use_namespaces
+        # is synonomous with use_neutron
+        # this is a test on whether access_instance has a floating_ip
+        # in the neutron case, we check the node group for the
+        # access_instance and look for fp
+        # in the nova case, we compare management_ip to internal_ip or even
+        # use the nova interface
+        elif (CONF.use_namespaces and not net_utils.has_floating_ip(
+                access_instance)):
             # need neutron info
             if not info:
                 info = self.get_neutron_info(access_instance)
@@ -743,37 +843,65 @@ class InstanceInteropHelper(remote.Remote):
 
     def execute_command(self, cmd, run_as_root=False, get_stderr=False,
                         raise_when_error=True, timeout=None):
-        self._log_command('Executing "%s"' % cmd)
-        return self._run_s(_execute_command, timeout, cmd, run_as_root,
-                           get_stderr, raise_when_error)
+        description = _('Executing "%s"') % cmd
+        self._log_command(description)
+        return self._run_s(_execute_command, timeout, description,
+                           cmd, run_as_root, get_stderr, raise_when_error)
 
     def write_file_to(self, remote_file, data, run_as_root=False,
                       timeout=None):
-        self._log_command('Writing file "%s"' % remote_file)
-        self._run_s(_write_file_to, timeout, remote_file, data, run_as_root)
+        description = _('Writing file "%s"') % remote_file
+        self._log_command(description)
+        self._run_s(_write_file_to, timeout, description,
+                    remote_file, data, run_as_root)
 
     def write_files_to(self, files, run_as_root=False, timeout=None):
-        self._log_command('Writing files "%s"' % list(files.keys()))
-        self._run_s(_write_files_to, timeout, files, run_as_root)
+        description = _('Writing files "%s"') % list(files)
+        self._log_command(description)
+        self._run_s(_write_files_to, timeout, description, files, run_as_root)
 
     def append_to_file(self, r_file, data, run_as_root=False, timeout=None):
-        self._log_command('Appending to file "%s"' % r_file)
-        self._run_s(_append_to_file, timeout, r_file, data, run_as_root)
+        description = _('Appending to file "%s"') % r_file
+        self._log_command(description)
+        self._run_s(_append_to_file, timeout, description,
+                    r_file, data, run_as_root)
 
     def append_to_files(self, files, run_as_root=False, timeout=None):
-        self._log_command('Appending to files "%s"' % list(files.keys()))
-        self._run_s(_append_to_files, timeout, files, run_as_root)
+        description = _('Appending to files "%s"') % list(files)
+        self._log_command(description)
+        self._run_s(_append_to_files, timeout, description, files, run_as_root)
 
     def read_file_from(self, remote_file, run_as_root=False, timeout=None):
-        self._log_command('Reading file "%s"' % remote_file)
-        return self._run_s(_read_file_from, timeout, remote_file, run_as_root)
+        description = _('Reading file "%s"') % remote_file
+        self._log_command(description)
+        return self._run_s(_read_file_from, timeout, description,
+                           remote_file, run_as_root)
+
+    def get_os_distrib(self, timeout=None):
+        return self._run_s(_get_os_distrib, timeout, "get_os_distrib")
+
+    def get_os_version(self, timeout=None):
+        return self._run_s(_get_os_version, timeout, "get_os_version")
+
+    def install_packages(self, packages, timeout=None):
+        description = _('Installing packages "%s"') % list(packages)
+        self._log_command(description)
+        self._run_s(_install_packages, timeout, description, packages)
+
+    def update_repository(self, timeout=None):
+        description = _('Updating repository')
+        self._log_command(description)
+        self._run_s(_update_repository, timeout, description)
 
     def replace_remote_string(self, remote_file, old_str, new_str,
                               timeout=None):
-        self._log_command('In file "%s" replacing string "%s" '
-                          'with "%s"' % (remote_file, old_str, new_str))
-        self._run_s(_replace_remote_string, timeout, remote_file, old_str,
-                    new_str)
+        description = _('In file "%(file)s" replacing string '
+                        '"%(old_string)s" with "%(new_string)s"') % {
+            "file": remote_file, "old_string": old_str, "new_string": new_str}
+
+        self._log_command(description)
+        self._run_s(_replace_remote_string, timeout, description,
+                    remote_file, old_str, new_str)
 
     def execute_on_vm_interactive(self, cmd, matcher, timeout=None):
         """Runs given command and responds to prompts.
@@ -789,9 +917,10 @@ class InstanceInteropHelper(remote.Remote):
              the command is finished. False should be returned
              otherwise.
         """
-
-        self._log_command('Executing interactively "%s"' % cmd)
-        self._run_s(_execute_on_vm_interactive, timeout, cmd, matcher)
+        description = _('Executing interactively "%s"') % cmd
+        self._log_command(description)
+        self._run_s(_execute_on_vm_interactive, timeout,
+                    description, cmd, matcher)
 
     def _log_command(self, str):
         with context.set_current_instance_id(self.instance.instance_id):
@@ -815,9 +944,9 @@ class BulkInstanceInteropHelper(InstanceInteropHelper):
     def _run(self, func, *args, **kwargs):
         return procutils.run_in_subprocess(self.proc, func, args, kwargs)
 
-    def _run_s(self, func, timeout, *args, **kwargs):
+    def _run_s(self, func, timeout, description, *args, **kwargs):
         timeout = _get_ssh_timeout(func, timeout)
-        return self._run_with_log(func, timeout, *args, **kwargs)
+        return self._run_with_log(func, timeout, description, *args, **kwargs)
 
 
 class SshRemoteDriver(remote.RemoteDriver):
